from types import FunctionType
from typing import Any, Dict, Optional, Tuple

valid_transforms: Dict[str, FunctionType] = {}

transform_list = [
    "PlateCarree",
    "LambertConformal",
    "LambertCylindrical",
    "Mercator",
    "Miller",
    "Mollweide",
    "Orthographic",
    "Robinson",
    "Stereographic",
    "TransverseMercator",
    "InterruptedGoodeHomolosine",
    "RotatedPole",
    "OSGB",
    "EuroPP",
    "Geostationary",
    "Gnomonic",
    "NorthPolarStereo",
    "OSNI",
    "SouthPolarStereo",
    "AlbersEqualArea",
    "AzimuthalEquidistant",
    "Sinusoidal",
    "UTM",
    "NearsidePerspective",
    "LambertAzimuthalEqualArea",
]


def cartopy_importer(transform_name):
    r"""Convenience function to import cartopy projection types"""

    def _func(*args, **kwargs):
        from yt.utilities.on_demand_imports import _cartopy as cartopy

        return getattr(cartopy.crs, transform_name)(*args, **kwargs)

    return _func


def get_mpl_transform(mpl_proj) -> Optional[FunctionType]:
    r"""This returns an instantiated transform function given a transform
    function name and arguments.

    Parameters
    ----------
    mpl_proj : string or tuple
        the matplotlib projection type. Can take the form of string or tuple.

    Examples
    --------

    >>> get_mpl_transform("PlateCarree")

    >>> get_mpl_transform(
    ...     ("RotatedPole", (), {"pole_latitude": 37.5, "pole_longitude": 177.5})
    ... )

    """
    # first check to see if the transform dict is empty, if it is fill it with
    # the cartopy functions
    if not valid_transforms:
        for mpl_transform in transform_list:
            valid_transforms[mpl_transform] = cartopy_importer(mpl_transform)

    # check to see if mpl_proj is a string or tuple, and construct args and
    # kwargs to pass to cartopy function based on that.
    key: Optional[str] = None
    args: Tuple = ()
    kwargs: Dict[str, Any] = {}
    if isinstance(mpl_proj, str):
        key = mpl_proj
        instantiated_func = valid_transforms[key](*args, **kwargs)
    elif isinstance(mpl_proj, tuple):
        if len(mpl_proj) == 2:
            key, args = mpl_proj
            kwargs = {}
        elif len(mpl_proj) == 3:
            key, args, kwargs = mpl_proj
        else:
            raise ValueError(f"Expected a tuple with len 2 or 3, received {mpl_proj}")
        if not isinstance(key, str):
            raise TypeError(
                f"Expected a string a the first element in mpl_proj, got {key!r}"
            )
        instantiated_func = valid_transforms[key](*args, **kwargs)
    elif hasattr(mpl_proj, "globe"):
        # cartopy transforms have a globe method associated with them
        key = mpl_proj
        instantiated_func = mpl_proj
    elif hasattr(mpl_proj, "set_transform"):
        # mpl axes objects have a set_transform method, so we'll check if that
        # exists on something passed in.
        key = mpl_proj
        instantiated_func = mpl_proj
    elif hasattr(mpl_proj, "name"):
        # last we'll check if the transform is in the list of registered
        # projections in matplotlib.
        from matplotlib.projections import get_projection_names

        registered_projections = get_projection_names()
        if mpl_proj.name in registered_projections:
            key = mpl_proj
            instantiated_func = mpl_proj
        else:
            key = None

    # build in a check that if none of the above options are satisfied by what
    # the user passes that None is returned for the instantiated function
    if key is None:
        instantiated_func = None
    return instantiated_func



import numpy as np

from yt.geometry.selection_routines import convert_mask_to_indices

from .grid_patch import AMRGridPatch


class StretchedGrid(AMRGridPatch):
    def __init__(self, id, cell_widths, *, filename=None, index=None):
        self.cell_widths = [np.array(_) for _ in cell_widths]
        super().__init__(id, filename, index)

    def _check_consistency(self):
        computed_right_edge = self.LeftEdge + [
            _.sum() for _ in self.cell_widths * self.LeftEdge.uq
        ]
        assert (computed_right_edge == self.RightEdge).all()

    def _get_selector_mask(self, selector):
        if self._cache_mask and hash(selector) == self._last_selector_id:
            mask = self._last_mask
        else:
            mask = selector.fill_mask(self)
            if self._cache_mask:
                self._last_mask = mask
            self._last_selector_id = hash(selector)
            if mask is None:
                self._last_count = 0
            else:
                self._last_count = mask.sum()
        return mask

    def select_fwidth(self, dobj):
        mask = self._get_selector_mask(dobj.selector)
        if mask is None:
            return np.empty((0, 3), dtype="float64")
        indices = convert_mask_to_indices(mask, self._last_count)

        coords = np.array(
            [
                self.cell_widths[0][indices[:, 0]],
                self.cell_widths[1][indices[:, 1]],
                self.cell_widths[2][indices[:, 2]],
            ]
        ).T
        return coords

    def select_fcoords(self, dobj):
        mask = self._get_selector_mask(dobj.selector)
        if mask is None:
            return np.empty((0, 3), dtype="float64")
        cell_centers = [
            self.LeftEdge[i].d
            + np.add.accumulate(self.cell_widths[i])
            - 0.5 * self.cell_widths[i]
            for i in range(3)
        ]
        indices = convert_mask_to_indices(mask, self._last_count)
        coords = np.array(
            [
                cell_centers[0][indices[:, 0]],
                cell_centers[1][indices[:, 1]],
                cell_centers[2][indices[:, 2]],
            ]
        ).T
        return coords



from nose.plugins.attrib import attr
from nose.tools import assert_raises

import yt
from yt.testing import ANSWER_TEST_TAG, assert_equal, fake_random_ds
from yt.utilities.answer_testing.framework import GenericImageTest
from yt.visualization.line_plot import _validate_point


def setup():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def compare(ds, plot, test_prefix, test_name, decimals=12):
    def image_from_plot(filename_prefix):
        return plot.save(filename_prefix)

    image_from_plot.__name__ = f"line_{test_prefix}"
    test = GenericImageTest(ds, image_from_plot, decimals)
    test.prefix = test_prefix
    test.answer_name = test_name
    return test


@attr(ANSWER_TEST_TAG)
def test_line_plot():
    ds = fake_random_ds(4)
    fields = [field for field in ds.field_list if field[0] == "stream"]
    field_labels = {f: f[1] for f in fields}
    plot = yt.LinePlot(
        ds, fields, (0, 0, 0), (1, 1, 0), 1000, field_labels=field_labels
    )
    plot.annotate_legend(fields[0])
    plot.annotate_legend(fields[1])
    plot.set_x_unit("cm")
    plot.set_unit(fields[0], "kg/cm**3")
    plot.annotate_title(fields[0], "Density Plot")
    yield compare(
        ds, plot, test_prefix="answers_line_plot", test_name="answers_line_plot"
    )


@attr(ANSWER_TEST_TAG)
def test_multi_line_plot():
    ds = fake_random_ds(4)
    fields = [field for field in ds.field_list if field[0] == "stream"]
    field_labels = {f: f[1] for f in fields}
    lines = []
    lines.append(yt.LineBuffer(ds, [0.25, 0, 0], [0.25, 1, 0], 100, label="x = 0.5"))
    lines.append(yt.LineBuffer(ds, [0.5, 0, 0], [0.5, 1, 0], 100, label="x = 0.5"))
    plot = yt.LinePlot.from_lines(ds, fields, lines, field_labels=field_labels)
    plot.annotate_legend(fields[0])
    plot.annotate_legend(fields[1])
    yield compare(
        ds,
        plot,
        test_prefix="answers_multi_line_plot",
        test_name="answers_multi_line_plot",
    )


def test_line_buffer():
    ds = fake_random_ds(32)
    lb = yt.LineBuffer(ds, (0, 0, 0), (1, 1, 1), 512, label="diag")
    lb[("gas", "density")]
    lb[("gas", "velocity_x")]
    assert_equal(lb[("gas", "density")].size, 512)
    lb[("gas", "density")] = 0
    assert_equal(lb[("gas", "density")], 0)
    assert_equal(set(lb.keys()), {("gas", "density"), ("gas", "velocity_x")})
    del lb[("gas", "velocity_x")]
    assert_equal(set(lb.keys()), {("gas", "density")})


def test_validate_point():
    ds = fake_random_ds(3)
    with assert_raises(RuntimeError) as ex:
        _validate_point(0, ds, start=True)
    assert_equal(str(ex.exception), "Input point must be array-like")

    with assert_raises(RuntimeError) as ex:
        _validate_point(ds.arr([[0], [1]], "code_length"), ds, start=True)
    assert_equal(str(ex.exception), "Input point must be a 1D array")

    with assert_raises(RuntimeError) as ex:
        _validate_point(ds.arr([0, 1], "code_length"), ds, start=True)
    assert_equal(
        str(ex.exception), "Input point must have an element for each dimension"
    )

    ds = fake_random_ds([32, 32, 1])
    _validate_point(ds.arr([0, 1], "code_length"), ds, start=True)
    _validate_point(ds.arr([0, 1], "code_length"), ds)



"""
Matching points on the grid to specific grids



"""


import numpy as np

cimport cython
cimport grid_visitors
cimport numpy as np
from grid_visitors cimport (
    GridTreeNode,
    GridTreeNodePadded,
    GridVisitorData,
    grid_visitor_function,
)
from libc.stdlib cimport free, malloc

from yt.geometry.selection_routines cimport SelectorObject, _ensure_code
from yt.utilities.lib.bitarray cimport bitarray
from yt.utilities.lib.fp_utils cimport iclip


cdef class GridTree:
    cdef GridTreeNode *grids
    cdef GridTreeNode *root_grids
    cdef int num_grids
    cdef int num_root_grids
    cdef int num_leaf_grids
    cdef public bitarray mask
    cdef void setup_data(self, GridVisitorData *data)
    cdef void visit_grids(self, GridVisitorData *data,
                          grid_visitor_function *func,
                          SelectorObject selector)
    cdef void recursively_visit_grid(self,
                          GridVisitorData *data,
                          grid_visitor_function *func,
                          SelectorObject selector,
                          GridTreeNode *grid,
                          np.uint8_t *buf = ?)

cdef class MatchPointsToGrids:

    cdef int num_points
    cdef np.float64_t *xp
    cdef np.float64_t *yp
    cdef np.float64_t *zp
    cdef GridTree tree
    cdef np.int64_t *point_grids
    cdef np.uint8_t check_position(self,
                                   np.int64_t pt_index,
                                   np.float64_t x,
                                   np.float64_t y,
                                   np.float64_t z,
                                   GridTreeNode *grid)

    cdef np.uint8_t is_in_grid(self,
			 np.float64_t x,
			 np.float64_t y,
			 np.float64_t z,
			 GridTreeNode *grid)

cdef extern from "platform_dep.h" nogil:
    double rint(double x)



import numpy as np

from yt.testing import assert_equal
from yt.utilities.answer_testing.framework import AnswerTestingTest


class ExtractConnectedSetsTest(AnswerTestingTest):
    _type_name = "ExtractConnectedSets"
    _attrs = ()

    def __init__(self, ds_fn, data_source, field, num_levels, min_val, max_val):
        super().__init__(ds_fn)
        self.data_source = data_source
        self.field = field
        self.num_levels = num_levels
        self.min_val = min_val
        self.max_val = max_val

    def run(self):
        n, all_sets = self.data_source.extract_connected_sets(
            self.field, self.num_levels, self.min_val, self.max_val
        )
        result = []
        for level in all_sets:
            for set_id in all_sets[level]:
                result.append(
                    [
                        all_sets[level][set_id]["cell_mass"].size,
                        all_sets[level][set_id]["cell_mass"].sum(),
                    ]
                )
        result = np.array(result)
        return result

    def compare(self, new_result, old_result):
        err_msg = f"Size and/or mass of connected sets do not agree for {self.ds_fn}."
        assert_equal(new_result, old_result, err_msg=err_msg, verbose=True)




# distutils: libraries = STD_LIBS
"""
Utilities for images
"""


import numpy as np

cimport cython
cimport numpy as np

from yt.utilities.lib.fp_utils cimport iclip


def add_points_to_greyscale_image(
        np.ndarray[np.float64_t, ndim=2] buffer,
        np.ndarray[np.uint8_t,   ndim=2] buffer_mask,
        np.ndarray[np.float64_t, ndim=1] px,
        np.ndarray[np.float64_t, ndim=1] py,
        np.ndarray[np.float64_t, ndim=1] pv):
    cdef int i, j, pi
    cdef int np = px.shape[0]
    cdef int xs = buffer.shape[0]
    cdef int ys = buffer.shape[1]
    for pi in range(np):
        j = <int> (xs * px[pi])
        i = <int> (ys * py[pi])
        buffer[i, j] += pv[pi]
        buffer_mask[i, j] = 1
    return

def add_points_to_image(
        np.ndarray[np.uint8_t, ndim=3] buffer,
        np.ndarray[np.float64_t, ndim=1] px,
        np.ndarray[np.float64_t, ndim=1] py,
        np.float64_t pv):
    cdef int i, j, k, pi
    cdef int np = px.shape[0]
    cdef int xs = buffer.shape[0]
    cdef int ys = buffer.shape[1]
    cdef int v
    v = iclip(<int>(pv * 255), 0, 255)
    for pi in range(np):
        j = <int> (xs * px[pi])
        i = <int> (ys * py[pi])
        for k in range(3):
            buffer[i, j, k] = v
        buffer[i, j, 3] = 255
    return

def add_rgba_points_to_image(
        np.ndarray[np.float64_t, ndim=3] buffer,
        np.ndarray[np.float64_t, ndim=1] px,
        np.ndarray[np.float64_t, ndim=1] py,
        np.ndarray[np.float64_t, ndim=2] rgba,
        ):
    """
    Splat rgba points onto an image

    Given an image buffer, add colors to
    pixels defined by fractional positions px and py,
    with colors rgba.  px and py are one dimensional
    arrays, and rgba is a an array of rgba values.
    """
    cdef int i, j, k, pi
    cdef int npart = px.shape[0]
    cdef int xs = buffer.shape[0]
    cdef int ys = buffer.shape[1]
    #iv = iclip(<int>(pv * 255), 0, 255)
    for pi in range(npart):
        j = <int> (xs * px[pi])
        i = <int> (ys * py[pi])
        if i < 0 or j < 0 or i >= xs or j >= ys:
            continue
        for k in range(4):
            buffer[i, j, k] += rgba[pi, k]
    return



"""
Generic file-handing functions for SPH data




"""
from yt.utilities.io_handler import BaseParticleIOHandler


class IOHandlerSPH(BaseParticleIOHandler):
    """IOHandler implementation specifically for SPH data

    This exists to handle particles with smoothing lengths, which require us
    to read in smoothing lengths along with the the particle coordinates to
    determine particle extents.

    At present this is non-functional.
    """

    pass



from nose.plugins.attrib import attr

from yt.testing import ANSWER_TEST_TAG, add_noise_fields, fake_random_ds
from yt.utilities.answer_testing.framework import GenericImageTest
from yt.visualization.api import PhasePlot


@attr(ANSWER_TEST_TAG)
def test_phaseplot_set_colorbar_properties_explicit():
    ds = fake_random_ds(16)
    add_noise_fields(ds)

    def create_image(filename_prefix):
        my_sphere = ds.sphere("c", 1)
        p = PhasePlot(
            my_sphere,
            ("gas", "noise1"),
            ("gas", "noise3"),
            [("gas", "density")],
            weight_field=None,
        )
        # using explicit units, we expect the colorbar units to stay unchanged
        p.set_zlim(("gas", "density"), zmin=(1e36, "kg/AU**3"))
        p.save(f"{filename_prefix}_set_zlim_explicit")

        # ... until we set them explicitly
        p.set_unit(("gas", "density"), "kg/AU**3")
        p.save(f"{filename_prefix}_set_zlim_set_unit_explicit")

    test = GenericImageTest(ds, create_image, 12)
    test.prefix = "test_phaseplot_set_colorbar_properties_explicit"
    test.answer_name = "phaseplot_set_colorbar_properties_explicit"
    yield test



from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
    YTSelectionContainer3D,
)
from yt.data_objects.static_output import Dataset
from yt.funcs import (
    validate_3d_array,
    validate_center,
    validate_object,
    validate_sequence,
)
from yt.units import YTArray


class YTRegion(YTSelectionContainer3D):
    """A 3D region of data with an arbitrary center.

    Takes an array of three *left_edge* coordinates, three
    *right_edge* coordinates, and a *center* that can be anywhere
    in the domain. If the selected region extends past the edges
    of the domain, no data will be found there, though the
    object's `left_edge` or `right_edge` are not modified.

    Parameters
    ----------
    center : array_like
        The center of the region
    left_edge : array_like
        The left edge of the region
    right_edge : array_like
        The right edge of the region
    """

    _type_name = "region"
    _con_args = ("center", "left_edge", "right_edge")

    def __init__(
        self,
        center,
        left_edge,
        right_edge,
        fields=None,
        ds=None,
        field_parameters=None,
        data_source=None,
    ):
        if center is not None:
            validate_center(center)
        validate_3d_array(left_edge)
        validate_3d_array(right_edge)
        validate_sequence(fields)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        YTSelectionContainer3D.__init__(self, center, ds, field_parameters, data_source)
        if not isinstance(left_edge, YTArray):
            self.left_edge = self.ds.arr(left_edge, "code_length", dtype="float64")
        else:
            # need to assign this dataset's unit registry to the YTArray
            self.left_edge = self.ds.arr(left_edge.copy(), dtype="float64")
        if not isinstance(right_edge, YTArray):
            self.right_edge = self.ds.arr(right_edge, "code_length", dtype="float64")
        else:
            # need to assign this dataset's unit registry to the YTArray
            self.right_edge = self.ds.arr(right_edge.copy(), dtype="float64")

    def _get_bbox(self):
        """
        Return the minimum bounding box for the region.
        """
        return self.left_edge.copy(), self.right_edge.copy()



import os
import shutil
import tempfile
from unittest import TestCase

import numpy as np

import yt
from yt.testing import fake_random_ds
from yt.visualization.volume_rendering.api import Scene, create_volume_source


def setup():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


class VariousVRTests(TestCase):
    # This toggles using a temporary directory. Turn off to examine images.
    use_tmpdir = True

    def setUp(self):
        if self.use_tmpdir:
            self.curdir = os.getcwd()
            # Perform I/O in safe place instead of yt main dir
            self.tmpdir = tempfile.mkdtemp()
            os.chdir(self.tmpdir)
        else:
            self.curdir, self.tmpdir = None, None

        self.ds = fake_random_ds(32)

    def tearDown(self):
        if self.use_tmpdir:
            os.chdir(self.curdir)
            shutil.rmtree(self.tmpdir)
        del self.ds

    def test_simple_scene_creation(self):
        yt.create_scene(self.ds)

    def test_modify_transfer_function(self):
        im, sc = yt.volume_render(self.ds)

        volume_source = sc.get_source(0)
        tf = volume_source.transfer_function
        tf.clear()
        tf.grey_opacity = True
        tf.add_layers(3, colormap="RdBu")
        sc.render()

    def test_multiple_fields(self):
        im, sc = yt.volume_render(self.ds)

        volume_source = sc.get_source(0)
        volume_source.set_field(("gas", "velocity_x"))
        volume_source.set_weight_field(("gas", "density"))
        sc.render()

    def test_rotation_volume_rendering(self):
        im, sc = yt.volume_render(self.ds)

        sc.camera.yaw(np.pi)
        sc.render()

    def test_simple_volume_rendering(self):
        im, sc = yt.volume_render(self.ds, sigma_clip=4.0)

    def test_lazy_volume_source_construction(self):
        sc = Scene()
        source = create_volume_source(self.ds.all_data(), ("gas", "density"))

        assert source._volume is None
        assert source._transfer_function is None

        source.tfh.bounds = (0.1, 1)

        source.set_log(False)

        assert not source.log_field
        assert source.transfer_function.x_bounds == [0.1, 1]
        assert source._volume is None

        source.set_log(True)

        assert source.log_field
        assert source.transfer_function.x_bounds == [-1, 0]
        assert source._volume is None

        source.transfer_function = None
        source.tfh.bounds = None

        ad = self.ds.all_data()

        np.testing.assert_allclose(
            source.transfer_function.x_bounds,
            np.log10(ad.quantities.extrema(("gas", "density"))),
        )
        assert source.tfh.log == source.log_field

        source.set_field(("gas", "velocity_x"))
        source.set_log(False)

        assert source.transfer_function.x_bounds == list(
            ad.quantities.extrema(("gas", "velocity_x"))
        )
        assert source._volume is None

        source.set_field(("gas", "density"))

        assert source.volume is not None
        assert not source.volume._initialized
        assert source.volume.fields is None

        del source.volume
        assert source._volume is None

        sc.add_source(source)

        sc.add_camera()

        sc.render()

        assert source.volume is not None
        assert source.volume._initialized
        assert source.volume.fields == [("gas", "density")]
        assert source.volume.log_fields == [True]

        source.set_field(("gas", "velocity_x"))
        source.set_log(False)

        sc.render()

        assert source.volume is not None
        assert source.volume._initialized
        assert source.volume.fields == [("gas", "velocity_x")]
        assert source.volume.log_fields == [False]



#
# ALL IMPORTS GO HERE
#

import os

import numpy as np

# This next item will handle most of the actual startup procedures, but it will
# also attempt to parse the command line and set up the global state of various
# operations.  The variable unparsed_args is not used internally but is
# provided as a convenience for users who wish to parse arguments in scripts.
# https://mail.python.org/archives/list/yt-dev@python.org/thread/L6AQPJ3OIMJC5SNKVM7CJG32YVQZRJWA/
import yt.startup_tasks as __startup_tasks
from yt import *
from yt._maintenance.deprecation import issue_deprecation_warning
from yt.config import ytcfg, ytcfg_defaults
from yt.utilities.logger import ytLogger

issue_deprecation_warning(
    "The yt.mods module is deprecated.", since="4.1.0", removal="4.2.0"
)

unparsed_args = __startup_tasks.unparsed_args


if ytLogger.getEffectiveLevel() >= int(ytcfg_defaults["yt"]["log_level"]):  # type: ignore
    # This won't get displayed.
    mylog.debug("Turning off NumPy error reporting")
    np.seterr(all="ignore")

# We load plugins.  Keep in mind, this can be fairly dangerous -
# the primary purpose is to allow people to have a set of functions
# that get used every time that they don't have to *define* every time.
# This way, other command-line tools can be used very simply.
# Unfortunately, for now, I think the easiest and simplest way of doing
# this is also the most dangerous way.
if ytcfg.get("yt", "load_field_plugins"):
    enable_plugins()



from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer

m_units = "Msun / h"  # Msun / h
p_units = "Mpccm / h"  # Mpc / h comoving
v_units = "km / s"  # km /s phys, peculiar
r_units = "kpccm / h"  # kpc / h comoving


class RockstarFieldInfo(FieldInfoContainer):

    known_particle_fields: KnownFieldsT = (
        ("particle_identifier", ("", [], None)),
        ("particle_position_x", (p_units, [], None)),
        ("particle_position_y", (p_units, [], None)),
        ("particle_position_z", (p_units, [], None)),
        ("particle_velocity_x", (v_units, [], None)),
        ("particle_velocity_y", (v_units, [], None)),
        ("particle_velocity_z", (v_units, [], None)),
        ("particle_corevel_x", (v_units, [], None)),
        ("particle_corevel_y", (v_units, [], None)),
        ("particle_corevel_z", (v_units, [], None)),
        ("particle_bulkvel_x", (v_units, [], None)),
        ("particle_bulkvel_y", (v_units, [], None)),
        ("particle_bulkvel_z", (v_units, [], None)),
        ("particle_mass", (m_units, [], "Mass")),
        ("virial_radius", (r_units, [], "Radius")),
        ("child_r", (r_units, [], None)),
        ("vmax_r", (v_units, [], None)),
        # These fields I don't have good definitions for yet.
        ("mgrav", ("", [], None)),
        ("vmax", (v_units, [], "V_{max}")),
        ("rvmax", (v_units, [], None)),
        ("rs", (r_units, [], "R_s")),
        ("klypin_rs", (r_units, [], "Klypin R_s")),
        ("vrms", (v_units, [], "V_{rms}")),
        ("Jx", ("", [], "J_x")),
        ("Jy", ("", [], "J_y")),
        ("Jz", ("", [], "J_z")),
        ("energy", ("", [], None)),
        ("spin", ("", [], "Spin Parameter")),
        ("alt_m1", (m_units, [], None)),
        ("alt_m2", (m_units, [], None)),
        ("alt_m3", (m_units, [], None)),
        ("alt_m4", (m_units, [], None)),
        ("Xoff", ("", [], None)),
        ("Voff", ("", [], None)),
        ("b_to_a", ("", [], "Ellipsoidal b to a")),
        ("c_to_a", ("", [], "Ellipsoidal c to a")),
        ("Ax", ("", [], "A_x")),
        ("Ay", ("", [], "A_y")),
        ("Az", ("", [], "A_z")),
        ("b_to_a2", ("", [], None)),
        ("c_to_a2", ("", [], None)),
        ("A2x", ("", [], "A2_x")),
        ("A2y", ("", [], "A2_y")),
        ("A2z", ("", [], "A2_z")),
        ("bullock_spin", ("", [], "Bullock Spin Parameter")),
        ("kin_to_pot", ("", [], "Kinetic to Potential")),
        ("m_pe_b", ("", [], None)),
        ("m_pe_d", ("", [], None)),
        ("num_p", ("", [], "Number of Particles")),
        ("num_child_particles", ("", [], "Number of Child Particles")),
        ("p_start", ("", [], None)),
        ("desc", ("", [], None)),
        ("flags", ("", [], None)),
        ("n_core", ("", [], None)),
        ("min_pos_err", ("", [], None)),
        ("min_vel_err", ("", [], None)),
        ("min_bulkvel_err", ("", [], None)),
    )



import numpy as np

from yt.testing import assert_almost_equal, fake_sph_orientation_ds, requires_module
from yt.utilities.lib.pixelization_routines import pixelize_sph_kernel_projection
from yt.utilities.on_demand_imports import _scipy
from yt.visualization.volume_rendering import off_axis_projection as OffAP

spatial = _scipy.spatial
ndimage = _scipy.ndimage


def test_no_rotation():
    """Determines if a projection processed through
    off_axis_projection with no rotation will give the same
    image buffer if processed directly through
    pixelize_sph_kernel_projection
    """
    normal_vector = [0.0, 0.0, 1.0]
    resolution = (64, 64)
    ds = fake_sph_orientation_ds()
    ad = ds.all_data()
    left_edge = ds.domain_left_edge
    right_edge = ds.domain_right_edge
    center = (left_edge + right_edge) / 2
    width = right_edge - left_edge
    px = ad[("all", "particle_position_x")]
    py = ad[("all", "particle_position_y")]
    hsml = ad[("all", "smoothing_length")]
    quantity_to_smooth = ad[("gas", "density")]
    density = ad[("io", "density")]
    mass = ad[("io", "particle_mass")]
    bounds = [-4, 4, -4, 4, -4, 4]

    buf2 = np.zeros(resolution)
    buf1 = OffAP.off_axis_projection(
        ds, center, normal_vector, width, resolution, ("gas", "density")
    )
    pixelize_sph_kernel_projection(
        buf2, px, py, hsml, mass, density, quantity_to_smooth, bounds
    )
    assert_almost_equal(buf1.ndarray_view(), buf2)


@requires_module("scipy")
def test_basic_rotation_1():
    """All particles on Z-axis should now be on the negative Y-Axis
    fake_sph_orientation has three z-axis particles,
    so there should be three y-axis particles after rotation
    (0, 0, 1) -> (0, -1)
    (0, 0, 2) -> (0, -2)
    (0, 0, 3) -> (0, -3)
    In addition, we should find a local maxima at (0, 0) due to:
    (0, 0, 0) -> (0, 0)
    (0, 1, 0) -> (0, 0)
    (0, 2, 0) -> (0, 0)
    and the one particle on the x-axis should not change its position:
    (1, 0, 0) -> (1, 0)
    """
    expected_maxima = ([0.0, 0.0, 0.0, 0.0, 1.0], [0.0, -1.0, -2.0, -3.0, 0.0])
    normal_vector = [0.0, 1.0, 0.0]
    north_vector = [0.0, 0.0, -1.0]
    resolution = (64, 64)
    ds = fake_sph_orientation_ds()
    left_edge = ds.domain_left_edge
    right_edge = ds.domain_right_edge
    center = (left_edge + right_edge) / 2
    width = right_edge - left_edge
    buf1 = OffAP.off_axis_projection(
        ds,
        center,
        normal_vector,
        width,
        resolution,
        ("gas", "density"),
        north_vector=north_vector,
    )
    find_compare_maxima(expected_maxima, buf1, resolution, width)


@requires_module("scipy")
def test_basic_rotation_2():
    """Rotation of x-axis onto z-axis.
    All particles on z-axis should now be on the negative x-Axis fake_sph_orientation
    has three z-axis particles, so there should be three x-axis particles after rotation
    (0, 0, 1) -> (-1, 0)
    (0, 0, 2) -> (-2, 0)
    (0, 0, 3) -> (-3, 0)
    In addition, we should find a local maxima at (0, 0) due to:
    (0, 0, 0) -> (0, 0)
    (1, 0, 0) -> (0, 0)
    and the two particles on the y-axis should not change its position:
    (0, 1, 0) -> (0, 1)
    (0, 2, 0) -> (0, 2)
    """
    expected_maxima = (
        [-1.0, -2.0, -3.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 1.0, 2.0],
    )
    normal_vector = [1.0, 0.0, 0.0]
    north_vector = [0.0, 1.0, 0.0]
    resolution = (64, 64)
    ds = fake_sph_orientation_ds()
    left_edge = ds.domain_left_edge
    right_edge = ds.domain_right_edge
    center = (left_edge + right_edge) / 2
    width = right_edge - left_edge
    buf1 = OffAP.off_axis_projection(
        ds,
        center,
        normal_vector,
        width,
        resolution,
        ("gas", "density"),
        north_vector=north_vector,
    )
    find_compare_maxima(expected_maxima, buf1, resolution, width)


@requires_module("scipy")
def test_basic_rotation_3():
    """Rotation of z-axis onto negative z-axis.
    All fake particles on z-axis should now be of the negative z-Axis.
    fake_sph_orientation has three z-axis particles,
    so we should have a local maxima at (0, 0)
    (0, 0, 1) -> (0, 0)
    (0, 0, 2) -> (0, 0)
    (0, 0, 3) -> (0, 0)
    In addition, (0, 0, 0) should also contribute to the local maxima at (0, 0):
    (0, 0, 0) -> (0, 0)
    x-axis particles should be rotated as such:
    (1, 0, 0) -> (0, -1)
    and same goes for y-axis particles:
    (0, 1, 0) -> (-1, 0)
    (0, 2, 0) -> (-2, 0)
    """
    expected_maxima = ([0.0, 0.0, -1.0, -2.0], [0.0, -1.0, 0.0, 0.0])
    normal_vector = [0.0, 0.0, -1.0]
    resolution = (64, 64)
    ds = fake_sph_orientation_ds()
    left_edge = ds.domain_left_edge
    right_edge = ds.domain_right_edge
    center = (left_edge + right_edge) / 2
    width = right_edge - left_edge
    buf1 = OffAP.off_axis_projection(
        ds, center, normal_vector, width, resolution, ("gas", "density")
    )
    find_compare_maxima(expected_maxima, buf1, resolution, width)


@requires_module("scipy")
def test_basic_rotation_4():
    """Rotation of x-axis to z-axis and original z-axis to y-axis with the use
    of the north_vector. All fake particles on z-axis should now be on the
    y-Axis.  All fake particles on the x-axis should now be on the z-axis, and
    all fake particles on the y-axis should now be on the x-axis.

    (0, 0, 1) -> (0, 1)
    (0, 0, 2) -> (0, 2)
    (0, 0, 3) -> (0, 3)
    In addition, (0, 0, 0) should contribute to the local maxima at (0, 0):
    (0, 0, 0) -> (0, 0)
    x-axis particles should be rotated and contribute to the local maxima at (0, 0):
    (1, 0, 0) -> (0, 0)
    and the y-axis particles shift into the positive x direction:
    (0, 1, 0) -> (1, 0)
    (0, 2, 0) -> (2, 0)
    """
    expected_maxima = ([0.0, 0.0, 0.0, 0.0, 1.0, 2.0], [1.0, 2.0, 3.0, 0.0, 0.0, 0.0])
    normal_vector = [1.0, 0.0, 0.0]
    north_vector = [0.0, 0.0, 1.0]
    resolution = (64, 64)
    ds = fake_sph_orientation_ds()
    left_edge = ds.domain_left_edge
    right_edge = ds.domain_right_edge
    center = (left_edge + right_edge) / 2
    width = right_edge - left_edge
    buf1 = OffAP.off_axis_projection(
        ds,
        center,
        normal_vector,
        width,
        resolution,
        ("gas", "density"),
        north_vector=north_vector,
    )
    find_compare_maxima(expected_maxima, buf1, resolution, width)


@requires_module("scipy")
def test_center_1():
    """Change the center to [0, 3, 0]
    Every point will be shifted by 3 in the y-domain
    With this, we should not be able to see any of the y-axis particles
    (0, 1, 0) -> (0, -2)
    (0, 2, 0) -> (0, -1)
    (0, 0, 1) -> (0, -3)
    (0, 0, 2) -> (0, -3)
    (0, 0, 3) -> (0, -3)
    (0, 0, 0) -> (0, -3)
    (1, 0, 0) -> (1, -3)
    """
    expected_maxima = ([0.0, 0.0, 0.0, 1.0], [-2.0, -1.0, -3.0, -3.0])
    normal_vector = [0.0, 0.0, 1.0]
    resolution = (64, 64)
    ds = fake_sph_orientation_ds()
    left_edge = ds.domain_left_edge
    right_edge = ds.domain_right_edge
    # center = [(left_edge[0] + right_edge[0])/2,
    #            left_edge[1],
    #           (left_edge[2] + right_edge[2])/2]
    center = [0.0, 3.0, 0.0]
    width = right_edge - left_edge
    buf1 = OffAP.off_axis_projection(
        ds, center, normal_vector, width, resolution, ("gas", "density")
    )
    find_compare_maxima(expected_maxima, buf1, resolution, width)


@requires_module("scipy")
def test_center_2():
    """Change the center to [0, -1, 0]
    Every point will be shifted by 1 in the y-domain
    With this, we should not be able to see any of the y-axis particles
    (0, 1, 0) -> (0, 2)
    (0, 2, 0) -> (0, 3)
    (0, 0, 1) -> (0, 1)
    (0, 0, 2) -> (0, 1)
    (0, 0, 3) -> (0, 1)
    (0, 0, 0) -> (0, 1)
    (1, 0, 0) -> (1, 1)
    """
    expected_maxima = ([0.0, 0.0, 0.0, 1.0], [2.0, 3.0, 1.0, 1.0])
    normal_vector = [0.0, 0.0, 1.0]
    resolution = (64, 64)
    ds = fake_sph_orientation_ds()
    left_edge = ds.domain_left_edge
    right_edge = ds.domain_right_edge
    center = [0.0, -1.0, 0.0]
    width = right_edge - left_edge
    buf1 = OffAP.off_axis_projection(
        ds, center, normal_vector, width, resolution, ("gas", "density")
    )
    find_compare_maxima(expected_maxima, buf1, resolution, width)


@requires_module("scipy")
def test_center_3():
    """Change the center to the left edge, or [0, -8, 0]
    Every point will be shifted by 8 in the y-domain
    With this, we should not be able to see anything !
    """
    expected_maxima = ([], [])
    normal_vector = [0.0, 0.0, 1.0]
    resolution = (64, 64)
    ds = fake_sph_orientation_ds()
    left_edge = ds.domain_left_edge
    right_edge = ds.domain_right_edge
    center = [0.0, -1.0, 0.0]
    width = [
        (right_edge[0] - left_edge[0]),
        left_edge[1],
        (right_edge[2] - left_edge[2]),
    ]
    buf1 = OffAP.off_axis_projection(
        ds, center, normal_vector, width, resolution, ("gas", "density")
    )
    find_compare_maxima(expected_maxima, buf1, resolution, width)


@requires_module("scipy")
def find_compare_maxima(expected_maxima, buf, resolution, width):
    buf_ndarray = buf.ndarray_view()
    max_filter_buf = ndimage.maximum_filter(buf_ndarray, size=5)
    maxima = np.isclose(max_filter_buf, buf_ndarray, rtol=1e-09)

    # ignore contributions from zones of no smoothing
    for i in range(len(maxima)):
        for j in range(len(maxima[i])):
            if np.isclose(buf_ndarray[i, j], 0.0, 1e-09):
                maxima[i, j] = False
    coords = ([], [])

    for i in range(len(maxima)):
        for j in range(len(maxima[i])):
            if maxima[i, j]:
                coords[0].append(i)
                coords[1].append(j)
    pixel_tolerance = 2.0
    x_scaling_factor = resolution[0] / width[0]
    y_scaling_factor = resolution[1] / width[1]
    for i in range(len(expected_maxima[0])):
        found_match = False
        for j in range(len(coords[0])):
            # normalize coordinates
            x_coord = coords[0][j]
            y_coord = coords[1][j]
            x_coord -= resolution[0] / 2
            y_coord -= resolution[1] / 2
            x_coord /= x_scaling_factor
            y_coord /= y_scaling_factor
            if (
                spatial.distance.euclidean(
                    [x_coord, y_coord], [expected_maxima[0][i], expected_maxima[1][i]]
                )
                < pixel_tolerance
            ):
                found_match = True
                break
        if found_match is not True:
            raise AssertionError
    pass



import matplotlib
from nose.plugins.attrib import attr
from packaging.version import Version

from yt.testing import ANSWER_TEST_TAG, fake_random_ds, skipif
from yt.utilities.answer_testing.framework import GenericImageTest
from yt.visualization.api import SlicePlot

MPL_VERSION = Version(matplotlib.__version__)


@skipif(
    MPL_VERSION < Version("3.2"),
    reason=f"TwoSlopeNorm requires MPL 3.2, we have {MPL_VERSION}",
)
@attr(ANSWER_TEST_TAG)
def test_sliceplot_custom_norm():
    from matplotlib.colors import TwoSlopeNorm

    ds = fake_random_ds(16)

    def create_image(filename_prefix):
        field = ("gas", "density")
        p = SlicePlot(ds, "z", field)
        p.set_norm(field, norm=(TwoSlopeNorm(vcenter=0, vmin=-0.5, vmax=1)))
        p.save(f"{filename_prefix}")

    test = GenericImageTest(ds, create_image, 12)
    test.prefix = "test_sliceplot_custom_norm"
    test.answer_name = "sliceplot_custom_norm"
    yield test



import numpy as np
from nose.plugins.attrib import attr

from yt.loaders import load_uniform_grid
from yt.testing import ANSWER_TEST_TAG
from yt.utilities.answer_testing.framework import GenericImageTest
from yt.visualization.api import SlicePlot


@attr(ANSWER_TEST_TAG)
def test_inf_and_finite_values_zlim():
    # see https://github.com/yt-project/yt/issues/3901
    shape = (32, 16, 1)
    a = np.ones(16)
    b = np.ones((32, 16))
    c = np.reshape(a * b, shape)

    # injecting an inf
    c[0, 0, 0] = np.inf

    data = {("gas", "density"): c}

    ds = load_uniform_grid(
        data,
        shape,
        bbox=np.array([[0, 1], [0, 1], [0, 1]]),
    )

    def create_image(filename_prefix):
        p = SlicePlot(ds, "z", ("gas", "density"))

        # setting zlim manually
        p.set_zlim(("gas", "density"), -10, 10)
        p.save(filename_prefix)

    test = GenericImageTest(ds, create_image, 12)
    test.prefix = "test_inf_and_finite_values_zlim"
    test.answer_name = "inf_and_finite_values_zlim"
    yield test



import numpy as np
from more_itertools import always_iterable

import yt.geometry
from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
    YTSelectionContainer3D,
)
from yt.data_objects.static_output import Dataset
from yt.funcs import validate_object, validate_sequence


class YTBooleanContainer(YTSelectionContainer3D):
    r"""
    This is a boolean operation, accepting AND, OR, XOR, and NOT for combining
    multiple data objects.

    This object is not designed to be created directly; it is designed to be
    created implicitly by using one of the bitwise operations (&, \|, ^, \~) on
    one or two other data objects.  These correspond to the appropriate boolean
    operations, and the resultant object can be nested.

    Parameters
    ----------
    op : string
        Can be AND, OR, XOR, NOT or NEG.
    dobj1 : yt.data_objects.selection_objects.data_selection_objects.
            YTSelectionContainer
        The first selection object
    dobj2 : yt.data_objects.selection_objects.base_objects.YTSelectionContainer
        The second object

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
    >>> sp = ds.sphere("c", 0.1)
    >>> dd = ds.r[:, :, :]
    >>> new_obj = sp ^ dd
    >>> print(new_obj.sum("cell_volume"), dd.sum("cell_volume") - sp.sum("cell_volume"))
    """

    _type_name = "bool"
    _con_args = ("op", "dobj1", "dobj2")

    def __init__(
        self, op, dobj1, dobj2, ds=None, field_parameters=None, data_source=None
    ):
        YTSelectionContainer3D.__init__(self, None, ds, field_parameters, data_source)
        self.op = op.upper()
        self.dobj1 = dobj1
        self.dobj2 = dobj2
        name = f"Boolean{self.op}Selector"
        sel_cls = getattr(yt.geometry.selection_routines, name)
        self._selector = sel_cls(self)

    def _get_bbox(self):
        le1, re1 = self.dobj1._get_bbox()
        if self.op == "NOT":
            return le1, re1
        else:
            le2, re2 = self.dobj2._get_bbox()
            return np.minimum(le1, le2), np.maximum(re1, re2)


class YTIntersectionContainer3D(YTSelectionContainer3D):
    """
    This is a more efficient method of selecting the intersection of multiple
    data selection objects.

    Creating one of these objects returns the intersection of all of the
    sub-objects; it is designed to be a faster method than chaining & ("and")
    operations to create a single, large intersection.

    Parameters
    ----------
    data_objects : Iterable of YTSelectionContainer
        The data objects to intersect

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> sp1 = ds.sphere((0.4, 0.5, 0.6), 0.15)
    >>> sp2 = ds.sphere((0.38, 0.51, 0.55), 0.1)
    >>> sp3 = ds.sphere((0.35, 0.5, 0.6), 0.15)
    >>> new_obj = ds.intersection((sp1, sp2, sp3))
    >>> print(new_obj.sum("cell_volume"))
    """

    _type_name = "intersection"
    _con_args = ("data_objects",)

    def __init__(self, data_objects, ds=None, field_parameters=None, data_source=None):
        validate_sequence(data_objects)
        for obj in data_objects:
            validate_object(obj, YTSelectionContainer)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        YTSelectionContainer3D.__init__(self, None, ds, field_parameters, data_source)
        self.data_objects = list(always_iterable(data_objects))


class YTDataObjectUnion(YTSelectionContainer3D):
    """
    This is a more efficient method of selecting the union of multiple
    data selection objects.

    Creating one of these objects returns the union of all of the sub-objects;
    it is designed to be a faster method than chaining | (or) operations to
    create a single, large union.

    Parameters
    ----------
    data_objects : Iterable of YTSelectionContainer
        The data objects to union

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
    >>> sp1 = ds.sphere((0.4, 0.5, 0.6), 0.1)
    >>> sp2 = ds.sphere((0.3, 0.5, 0.15), 0.1)
    >>> sp3 = ds.sphere((0.5, 0.5, 0.9), 0.1)
    >>> new_obj = ds.union((sp1, sp2, sp3))
    >>> print(new_obj.sum("cell_volume"))
    """

    _type_name = "union"
    _con_args = ("data_objects",)

    def __init__(self, data_objects, ds=None, field_parameters=None, data_source=None):
        validate_sequence(data_objects)
        for obj in data_objects:
            validate_object(obj, YTSelectionContainer)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        YTSelectionContainer3D.__init__(self, None, ds, field_parameters, data_source)
        self.data_objects = list(always_iterable(data_objects))



cdef class DiskSelector(SelectorObject):
    cdef public np.float64_t norm_vec[3]
    cdef public np.float64_t center[3]
    cdef public np.float64_t radius, radius2
    cdef public np.float64_t height

    def __init__(self, dobj):
        cdef int i
        for i in range(3):
            self.norm_vec[i] = dobj._norm_vec[i]
            self.center[i] = _ensure_code(dobj.center[i])
        self.radius = _ensure_code(dobj.radius)
        self.radius2 = self.radius * self.radius
        self.height = _ensure_code(dobj.height)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
        return self.select_point(pos)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_point(self, np.float64_t pos[3]) nogil:
        cdef np.float64_t h, d, r2, temp
        cdef int i
        h = d = 0
        for i in range(3):
            temp = self.periodic_difference(pos[i], self.center[i], i)
            h += temp * self.norm_vec[i]
            d += temp*temp
        r2 = (d - h*h)
        if fabs(h) <= self.height and r2 <= self.radius2: return 1
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
        cdef np.float64_t h, d, r2, temp
        cdef int i
        h = d = 0
        for i in range(3):
            temp = self.periodic_difference(pos[i], self.center[i], i)
            h += temp * self.norm_vec[i]
            d += temp*temp
        r2 = (d - h*h)
        d = self.radius+radius
        if fabs(h) <= self.height+radius and r2 <= d*d: return 1
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) nogil:
        # Until we can get our OBB/OBB intersection correct, disable this.
        return 1
        # cdef np.float64_t *arr[2]
        # cdef np.float64_t pos[3]
        # cdef np.float64_t H, D, R2, temp
        # cdef int i, j, k, n
        # cdef int all_under = 1
        # cdef int all_over = 1
        # cdef int any_radius = 0
        # # A moment of explanation (revised):
        # #    The disk and bounding box collide if any of the following are true:
        # #    1) the center of the disk is inside the bounding box
        # #    2) any corner of the box lies inside the disk
        # #    3) the box spans the plane (!all_under and !all_over) and at least
        # #       one corner is within the cylindrical radius

        # # check if disk center lies inside bbox
        # if left_edge[0] <= self.center[0] <= right_edge[0] and \
        #    left_edge[1] <= self.center[1] <= right_edge[1] and \
        #    left_edge[2] <= self.center[2] <= right_edge[2] :
        #     return 1

        # # check all corners
        # arr[0] = left_edge
        # arr[1] = right_edge
        # for i in range(2):
        #     pos[0] = arr[i][0]
        #     for j in range(2):
        #         pos[1] = arr[j][1]
        #         for k in range(2):
        #             pos[2] = arr[k][2]
        #             H = D = 0
        #             for n in range(3):
        #                 temp = self.difference(pos[n], self.center[n], n)
        #                 H += (temp * self.norm_vec[n])
        #                 D += temp*temp
        #             R2 = (D - H*H)
        #             if R2 < self.radius2 :
        #                 any_radius = 1
        #                 if fabs(H) < self.height: return 1
        #             if H < 0: all_over = 0
        #             if H > 0: all_under = 0
        # if all_over == 0 and all_under == 0 and any_radius == 1: return 1
        # return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) nogil:
        # Until we can get our OBB/OBB intersection correct, disable this.
        return 2
        # cdef np.float64_t *arr[2]
        # cdef np.float64_t pos[3], H, D, R2, temp
        # cdef int i, j, k, n
        # cdef int all_under = 1
        # cdef int all_over = 1
        # cdef int any_radius = 0
        # # A moment of explanation (revised):
        # #    The disk and bounding box collide if any of the following are true:
        # #    1) the center of the disk is inside the bounding box
        # #    2) any corner of the box lies inside the disk
        # #    3) the box spans the plane (!all_under and !all_over) and at least
        # #       one corner is within the cylindrical radius

        # # check if disk center lies inside bbox
        # if left_edge[0] <= self.center[0] <= right_edge[0] and \
        #    left_edge[1] <= self.center[1] <= right_edge[1] and \
        #    left_edge[2] <= self.center[2] <= right_edge[2] :
        #     return 1

        # # check all corners
        # arr[0] = left_edge
        # arr[1] = right_edge
        # for i in range(2):
        #     pos[0] = arr[i][0]
        #     for j in range(2):
        #         pos[1] = arr[j][1]
        #         for k in range(2):
        #             pos[2] = arr[k][2]
        #             H = D = 0
        #             for n in range(3):
        #                 temp = self.periodic_difference(
        #                     pos[n], self.center[n], n)
        #                 H += (temp * self.norm_vec[n])
        #                 D += temp*temp
        #             R2 = (D - H*H)
        #             if R2 < self.radius2 :
        #                 any_radius = 1
        #                 if fabs(H) < self.height: return 1
        #             if H < 0: all_over = 0
        #             if H > 0: all_under = 0
        # if all_over == 0 and all_under == 0 and any_radius == 1: return 1
        # return 0

    def _hash_vals(self):
        return (("norm_vec[0]", self.norm_vec[0]),
                ("norm_vec[1]", self.norm_vec[1]),
                ("norm_vec[2]", self.norm_vec[2]),
                ("center[0]", self.center[0]),
                ("center[1]", self.center[1]),
                ("center[2]", self.center[2]),
                ("radius", self.radius),
                ("radius2", self.radius2),
                ("height", self.height))

    def _get_state_attnames(self):
        return ("radius", "radius2", "height", "norm_vec", "center")


disk_selector = DiskSelector



from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer
from yt.utilities.physical_constants import kboltz, mh

b_units = "code_magnetic"
pres_units = "code_mass/(code_length*code_time**2)"
rho_units = "code_mass / code_length**3"
vel_units = "code_length / code_time"


def velocity_field(j):
    def _velocity(field, data):
        return data["athena_pp", f"mom{j}"] / data["athena_pp", "dens"]

    return _velocity


class AthenaPPFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("rho", (rho_units, ["density"], None)),
        ("dens", (rho_units, ["density"], None)),
        ("Bcc1", (b_units, [], None)),
        ("Bcc2", (b_units, [], None)),
        ("Bcc3", (b_units, [], None)),
    )

    def setup_fluid_fields(self):
        from yt.fields.magnetic_field import setup_magnetic_field_aliases

        unit_system = self.ds.unit_system
        # Add velocity fields
        vel_prefix = "velocity"
        for i, comp in enumerate(self.ds.coordinates.axis_order):
            vel_field = ("athena_pp", "vel%d" % (i + 1))
            mom_field = ("athena_pp", "mom%d" % (i + 1))
            if vel_field in self.field_list:
                self.add_output_field(
                    vel_field, sampling_type="cell", units="code_length/code_time"
                )
                self.alias(
                    ("gas", f"{vel_prefix}_{comp}"),
                    vel_field,
                    units=unit_system["velocity"],
                )
            elif mom_field in self.field_list:
                self.add_output_field(
                    mom_field,
                    sampling_type="cell",
                    units="code_mass/code_time/code_length**2",
                )
                self.add_field(
                    ("gas", f"{vel_prefix}_{comp}"),
                    sampling_type="cell",
                    function=velocity_field(i + 1),
                    units=unit_system["velocity"],
                )
        # Figure out thermal energy field
        if ("athena_pp", "press") in self.field_list:
            self.add_output_field(
                ("athena_pp", "press"), sampling_type="cell", units=pres_units
            )
            self.alias(
                ("gas", "pressure"),
                ("athena_pp", "press"),
                units=unit_system["pressure"],
            )

            def _specific_thermal_energy(field, data):
                return (
                    data["athena_pp", "press"]
                    / (data.ds.gamma - 1.0)
                    / data["athena_pp", "rho"]
                )

            self.add_field(
                ("gas", "specific_thermal_energy"),
                sampling_type="cell",
                function=_specific_thermal_energy,
                units=unit_system["specific_energy"],
            )
        elif ("athena_pp", "Etot") in self.field_list:
            self.add_output_field(
                ("athena_pp", "Etot"), sampling_type="cell", units=pres_units
            )

            def _specific_thermal_energy(field, data):
                eint = data["athena_pp", "Etot"] - data["gas", "kinetic_energy_density"]
                if ("athena_pp", "B1") in self.field_list:
                    eint -= data["gas", "magnetic_energy_density"]
                return eint / data["athena_pp", "dens"]

            self.add_field(
                ("gas", "specific_thermal_energy"),
                sampling_type="cell",
                function=_specific_thermal_energy,
                units=unit_system["specific_energy"],
            )

        # Add temperature field
        def _temperature(field, data):
            return (
                (data["gas", "pressure"] / data["gas", "density"])
                * data.ds.mu
                * mh
                / kboltz
            )

        self.add_field(
            ("gas", "temperature"),
            sampling_type="cell",
            function=_temperature,
            units=unit_system["temperature"],
        )

        setup_magnetic_field_aliases(
            self, "athena_pp", ["Bcc%d" % ax for ax in (1, 2, 3)]
        )



import numpy as np

from yt.funcs import mylog
from yt.utilities.io_handler import BaseParticleIOHandler
from yt.utilities.on_demand_imports import _h5py as h5py

_pos_names = ["CenterOfMass", "CentreOfMass"]


class IOHandlerOWLSSubfindHDF5(BaseParticleIOHandler):
    _dataset_type = "subfind_hdf5"
    _position_name = None

    def __init__(self, ds):
        super().__init__(ds)
        self.offset_fields = set()

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    def _read_particle_coords(self, chunks, ptf):
        # This will read chunks and yield the results.
        chunks = list(chunks)
        data_files = set()
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files, key=lambda x: (x.filename, x.start)):
            with h5py.File(data_file.filename, mode="r") as f:
                for ptype in sorted(ptf):
                    pcount = data_file.total_particles[ptype]
                    coords = f[ptype][self._position_name][()].astype("float64")
                    coords = np.resize(coords, (pcount, 3))
                    x = coords[:, 0]
                    y = coords[:, 1]
                    z = coords[:, 2]
                    yield ptype, (x, y, z), 0.0

    def _yield_coordinates(self, data_file):
        ptypes = self.ds.particle_types_raw
        with h5py.File(data_file.filename, mode="r") as f:
            for ptype in sorted(ptypes):
                pcount = data_file.total_particles[ptype]
                coords = f[ptype][self._position_name][()].astype("float64")
                coords = np.resize(coords, (pcount, 3))
                yield ptype, coords

    def _read_offset_particle_field(self, field, data_file, fh):
        field_data = np.empty(data_file.total_particles["FOF"], dtype="float64")
        fofindex = (
            np.arange(data_file.total_particles["FOF"]) + data_file.index_start["FOF"]
        )
        for offset_file in data_file.offset_files:
            if fh.filename == offset_file.filename:
                ofh = fh
            else:
                ofh = h5py.File(offset_file.filename, mode="r")
            subindex = np.arange(offset_file.total_offset) + offset_file.offset_start
            substart = max(fofindex[0] - subindex[0], 0)
            subend = min(fofindex[-1] - subindex[0], subindex.size - 1)
            fofstart = substart + subindex[0] - fofindex[0]
            fofend = subend + subindex[0] - fofindex[0]
            field_data[fofstart : fofend + 1] = ofh["SUBFIND"][field][
                substart : subend + 1
            ]
        return field_data

    def _read_particle_fields(self, chunks, ptf, selector):
        # Now we have all the sizes, and we can allocate
        chunks = list(chunks)
        data_files = set()
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files, key=lambda x: (x.filename, x.start)):
            with h5py.File(data_file.filename, mode="r") as f:
                for ptype, field_list in sorted(ptf.items()):
                    pcount = data_file.total_particles[ptype]
                    if pcount == 0:
                        continue
                    coords = f[ptype][self._position_name][()].astype("float64")
                    coords = np.resize(coords, (pcount, 3))
                    x = coords[:, 0]
                    y = coords[:, 1]
                    z = coords[:, 2]
                    mask = selector.select_points(x, y, z, 0.0)
                    del x, y, z
                    if mask is None:
                        continue
                    for field in field_list:
                        if field in self.offset_fields:
                            field_data = self._read_offset_particle_field(
                                field, data_file, f
                            )
                        else:
                            if field == "particle_identifier":
                                field_data = (
                                    np.arange(data_file.total_particles[ptype])
                                    + data_file.index_start[ptype]
                                )
                            elif field in f[ptype]:
                                field_data = f[ptype][field][()].astype("float64")
                            else:
                                fname = field[: field.rfind("_")]
                                field_data = f[ptype][fname][()].astype("float64")
                                my_div = field_data.size / pcount
                                if my_div > 1:
                                    field_data = np.resize(
                                        field_data, (int(pcount), int(my_div))
                                    )
                                    findex = int(field[field.rfind("_") + 1 :])
                                    field_data = field_data[:, findex]
                        data = field_data[mask]
                        yield (ptype, field), data

    def _count_particles(self, data_file):
        with h5py.File(data_file.filename, mode="r") as f:
            pcount = {"FOF": get_one_attr(f["FOF"], ["Number_of_groups", "Ngroups"])}
            if "SUBFIND" in f:
                # We need this to figure out where the offset fields are stored.
                data_file.total_offset = get_one_attr(
                    f["SUBFIND"], ["Number_of_groups", "Ngroups"]
                )
                pcount["SUBFIND"] = get_one_attr(
                    f["FOF"], ["Number_of_subgroups", "Nsubgroups"]
                )
            else:
                data_file.total_offset = 0
                pcount["SUBFIND"] = 0

        return pcount

    def _identify_fields(self, data_file):
        fields = []
        pcount = data_file.total_particles
        if sum(pcount.values()) == 0:
            return fields, {}
        with h5py.File(data_file.filename, mode="r") as f:
            for ptype in self.ds.particle_types_raw:
                if data_file.total_particles[ptype] == 0:
                    continue
                self._identify_position_name(f[ptype])
                fields.append((ptype, "particle_identifier"))
                my_fields, my_offset_fields = subfind_field_list(
                    f[ptype], ptype, data_file.total_particles
                )
                fields.extend(my_fields)
                self.offset_fields = self.offset_fields.union(set(my_offset_fields))
        return fields, {}

    def _identify_position_name(self, fh):
        if self._position_name is not None:
            return
        for pname in _pos_names:
            if pname in fh:
                self._position_name = pname
                return


def get_one_attr(fh, attrs, default=None, error=True):
    """
    Try getting from a list of attrs. Return the first one that exists.
    """
    for attr in attrs:
        if attr in fh.attrs:
            return fh.attrs[attr]
    if error:
        raise RuntimeError(
            f"Could not find any of these attributes: {attrs}. "
            f"Available attributes: {fh.attrs.keys()}"
        )
    return default


def subfind_field_list(fh, ptype, pcount):
    fields = []
    offset_fields = []
    for field in fh.keys():
        if "PartType" in field:
            # These are halo member particles
            continue
        elif isinstance(fh[field], h5py.Group):
            my_fields, my_offset_fields = subfind_field_list(fh[field], ptype, pcount)
            fields.extend(my_fields)
            my_offset_fields.extend(offset_fields)
        else:
            if not fh[field].size % pcount[ptype]:
                my_div = fh[field].size / pcount[ptype]
                fname = fh[field].name[fh[field].name.find(ptype) + len(ptype) + 1 :]
                if my_div > 1:
                    for i in range(int(my_div)):
                        fields.append((ptype, "%s_%d" % (fname, i)))
                else:
                    fields.append((ptype, fname))
            elif (
                ptype == "SUBFIND"
                and not fh[field].size % fh["/SUBFIND"].attrs["Number_of_groups"]
            ):
                # These are actually FOF fields, but they were written after
                # a load balancing step moved halos around and thus they do not
                # correspond to the halos stored in the FOF group.
                my_div = fh[field].size / fh["/SUBFIND"].attrs["Number_of_groups"]
                fname = fh[field].name[fh[field].name.find(ptype) + len(ptype) + 1 :]
                if my_div > 1:
                    for i in range(int(my_div)):
                        fields.append(("FOF", "%s_%d" % (fname, i)))
                else:
                    fields.append(("FOF", fname))
                offset_fields.append(fname)
            else:
                mylog.warning(
                    "Cannot add field (%s, %s) with size %d.",
                    ptype,
                    fh[field].name,
                    fh[field].size,
                )
                continue
    return fields, offset_fields
